// This file is part of ESTmapper.
// Copyright (c) 2001, 2002, 2003, 2004 Applera Corporation
// Authors: Brian Walenz and Liliana Florea
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received (LICENSE.txt) a copy of the GNU General Public 
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <string.h>

#include "bri++.H"

//
//  N.B. any read/write pair (either way) must have a seek (or a fflush) in between.
//

bitPackedFile::bitPackedFile(char const *name) {

  //  Try to open the original name -- we don't support compressed
  //  files for rewrite.  We just fail with a can't open message.
  //
  //  To get read/write and create we have to use open(2), as mode
  //  "r+" of fopen(3) will not create.
  //
  errno = 0;
  _file = open(name, O_RDWR | O_CREAT | O_LARGEFILE,
               S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
  int errno_fopen = errno;

  //  If that fails, try to open the compressed version
  //
  if (errno_fopen) {
    size_t  l = strlen(name);

    char *command = new char [l + 64];
    sprintf(command, "%s.bz2", name);

    if (fileExists(command)) {
      fprintf(stderr, "bitPackedFile::bitPackedFile()-- Can't use compressed bitPackedFiles here.\n");
      fprintf(stderr, "bitPackedFile::bitPackedFile()-- '%s' not opened.\n", command);
    } else {
      fprintf(stderr, "bitPackedFile::bitPackedFile()-- Couldn't open bitPackedFile.\n");
      fprintf(stderr, "bitPackedFile::bitPackedFile()-- %s: %s\n", name, strerror(errno_fopen));
    }

    exit(1);
  }

  _bfrmax = 4096;
  _bfrmax = 1048576 / 8;
  _bfr    = new u64bit [_bfrmax];
  _pos    = u64bitZERO;
  _bit    = u64bitZERO;

  _bfrDirty = false;

  stat_seekInside   = u64bitZERO;
  stat_seekOutside  = u64bitZERO;
  stat_dirtyFlushes = u64bitZERO;

  seek(0, true);
}


bitPackedFile::~bitPackedFile() {
  flushDirty();
  delete [] _bfr;
  close(_file);
}




//  If the page is dirty, flush it to disk
//
void
bitPackedFile::flushDirty(void) {

  if (_bfrDirty) {
    stat_dirtyFlushes++;

    errno = 0;
    lseek(_file, _pos * sizeof(u64bit), SEEK_SET);
    if (errno) {
      fprintf(stderr, "bitPackedFile::seek() failed: %s\n", strerror(errno));
      exit(1);
    }

    //  We should only write bits up to _bit, the position we are
    //  currently at.  However, we don't know if the block is being
    //  flushed because we're totally finished with it, or because we
    //  are moving on to the next block.  If we're done with it, we
    //  want to flush the word that contains _bit, and if we're moving
    //  on to the next one, we'll flush that word again.  So, in
    //  either case, we flush the word that contains _bit.
    //
    errno = 0;
    write(_file, _bfr, sizeof(u64bit) * _bfrmax);
    if (errno) {
      fprintf(stderr, "bitPackedFile::write() failed: %s\n", strerror(errno));
      exit(1);
    }

    _bfrDirty = false;
  }
}





//  Seeks to bitposition pos in the file, reads in a new block.
//
void
bitPackedFile::seek(u64bit bitpos, bool forceLoad) {

  //  If we are seeking to somewhere in the current block, don't do a
  //  real seek, just move our position within the block.
  //
  if (forceLoad == false) {
    u64bit np = bitpos >> 6;
    
    if ((_pos <= np) && (np <= _pos + _bfrmax - 32)) {
      _bit = bitpos - (_pos << 6);
      stat_seekInside++;
      //fprintf(stderr, "SEEK INSIDE to _bit="u64bitFMT"\n", _bit);
      return;
    }
  }

  stat_seekOutside++;

  flushDirty();

  //  Somewhat of a gross hack to allow sequential access backwards.
  //
  //  If the new position (bitpos >> 6) is just before the old
  //  position (_pos), assume that we are being accessed iteratively
  //  backwards and load a full buffer so that the position we want to
  //  access is at the end.
  //
  //  Easy to think of bone-headed ways to break this (e.g., seek to
  //  the second element in a structure, access the first, then access
  //  the third).  Not so easy to think of a logical reason someone
  //  would want to do that.
  //
  if (((bitpos >> 6) < _pos) && ((bitpos >> 6) + 32 >= _pos)) {
    _pos = bitpos >> 6;
    if (_pos > _bfrmax)
      _pos = _pos - _bfrmax + 32;
    else
      _pos = 0;
  } else {
    _pos = bitpos >> 6;
  }

  _bit = bitpos - (_pos << 6);


  errno = 0;
  lseek(_file, _pos * 8, SEEK_SET);
  if (errno) {
    fprintf(stderr, "bitPackedFile::seek() failed: %s\n", strerror(errno));
    exit(1);
  }

  errno = 0;
  size_t wordsread = read(_file, _bfr, sizeof(u64bit) * _bfrmax);
  if (errno) {
    fprintf(stderr, "bitPackedFile::bitPackedFile got %s\n", strerror(errno));
    exit(1);
  }

  //  Clear any words that we didn't read (supposedly, because we hit
  //  EOF).
  //
  while (wordsread < _bfrmax)
    _bfr[wordsread++] = u64bitZERO;

  //fprintf(stderr, "SEEK OUTSIDE to _pos="u64bitFMT" _bit="u64bitFMT"\n", _pos, _bit);
}




u64bit
bitPackedFile::getBits(u32bit siz) {
  sync();
  u64bit ret = getDecodedValue(_bfr, _bit, siz);
  _bit += siz;
  return(ret);
}

void
bitPackedFile::putBits(u64bit bits, u32bit siz) {
  sync();
  setDecodedValue(_bfr, _bit, siz, bits);
  _bit += siz;
  _bfrDirty = true;
}

u64bit
bitPackedFile::getNumber(void) {
  sync();
  u64bit siz = 0;
  u64bit ret = getFibonacciEncodedNumber(_bfr, _bit, &siz);
  _bit += siz;
  return(ret);
}

void
bitPackedFile::putNumber(u64bit val) {
  sync();
  u64bit siz = 0;
  setFibonacciEncodedNumber(_bfr, _bit, &siz, val);
  _bit += siz;
  _bfrDirty = true;
}
