// This file is part of ESTmapper.
// Copyright (c) 2001, 2002, 2003, 2004 Applera Corporation
// Authors: Brian Walenz and Liliana Florea
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received (LICENSE.txt) a copy of the GNU General Public 
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "meryl.H"
#include "libmeryl.H"


void
binaryOperations(merylArgs *args) {

  if (args->mergeFilesLen != 2) {
    fprintf(stderr, "ERROR - must have exactly two files!\n");
    exit(1);
  }
  if (args->outputFile == 0L) {
    fprintf(stderr, "ERROR - no output file specified.\n");
    exit(1);
  }
  if ((args->personality != PERSONALITY_SUB) &&
      (args->personality != PERSONALITY_ABS)) {
    fprintf(stderr, "ERROR - only personalities sub and abs\n");
    fprintf(stderr, "ERROR - are supported in binaryOperations().\n");
    fprintf(stderr, "ERROR - this is a coding error, not a user error.\n");
    exit(1);
  }

  //  Open the input files, read in the first mer
  //
  merylStreamReader *A = new merylStreamReader(args->mergeFiles[0]);
  merylStreamReader *B = new merylStreamReader(args->mergeFiles[1]);

  A->nextMer();
  B->nextMer();

  //  Make sure that the mersizes agree, and pick a prefix size for
  //  the output
  //
  if (A->merSize() != B->merSize()) {
    fprintf(stderr, "ERROR - mersizes are different!\n");
    fprintf(stderr, "ERROR - mersize of '%s' is "u32bitFMT"\n", args->mergeFiles[0], A->merSize());
    fprintf(stderr, "ERROR - mersize of '%s' is "u32bitFMT"\n", args->mergeFiles[1], B->merSize());
    exit(1);
  }

  //  Open the output file, using the larger of the two prefix sizes
  //
  merylStreamWriter *W = new merylStreamWriter(args->outputFile,
                                               A->merSize(),
                                               (A->prefixSize() > B->prefixSize()) ? A->prefixSize() : B->prefixSize());




  //  SUB - report A - B
  //  ABS - report the absolute difference between the two files
  //
  //  These two operations are very similar (SUB was derived from ABS), so
  //  any bug found in one is probably in the other.
  //
  u64bit  Amer = u64bitZERO;
  u32bit  Acnt = u32bitZERO;
  u64bit  Bmer = u64bitZERO;
  u32bit  Bcnt = u32bitZERO;

  switch (args->personality) {
    case PERSONALITY_SUB:
      while (A->validMer() || B->validMer()) {
        Amer = A->theFMer();
        Acnt = A->theCount();
        Bmer = B->theFMer();
        Bcnt = B->theCount();

        //  If the A stream is all out of mers, set Amer to be the
        //  same as Bmer, and set Acnt to zero.  Similar for B.
        //
        if (!A->validMer()) {
          Amer = Bmer;
          Acnt = u32bitZERO;
        }
        if (!B->validMer()) {
          Bmer = Amer;
          Bcnt = u32bitZERO;
        }

        fprintf(stderr, "sub A="u64bitHEX" B="u64bitHEX"\n", Amer, Bmer);

        if (Amer == Bmer) {
          W->addMer(Amer, (Acnt > Bcnt) ? Acnt - Bcnt : 0);
          A->nextMer();
          B->nextMer();
        } else if (Amer < Bmer) {
          W->addMer(Amer, Acnt);
          A->nextMer();
        } else {
          B->nextMer();
        }
      }
      break;
    case PERSONALITY_ABS:
      while (A->validMer() || B->validMer()) {
        Amer = A->theFMer();
        Acnt = A->theCount();
        Bmer = B->theFMer();
        Bcnt = B->theCount();

        //  If the A stream is all out of mers, set Amer to be the
        //  same as Bmer, and set Acnt to zero.  Similar for B.
        //
        if (!A->validMer()) {
          Amer = Bmer;
          Acnt = u32bitZERO;
        }
        if (!B->validMer()) {
          Bmer = Amer;
          Bcnt = u32bitZERO;
        }

        if (Amer == Bmer) {
          W->addMer(Amer, (Acnt > Bcnt) ? Acnt - Bcnt : Bcnt - Acnt);
          A->nextMer();
          B->nextMer();
        } else if (Amer < Bmer) {
          W->addMer(Amer, Acnt);
          A->nextMer();
        } else {
          W->addMer(Bmer, Bcnt);
          B->nextMer();
        }
      }
      break;
  }

  delete A;
  delete B;
  delete W;
}
