// This file is part of ESTmapper.
// Copyright (c) 2001, 2002, 2003, 2004 Applera Corporation
// Authors: Brian Walenz and Liliana Florea
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received (LICENSE.txt) a copy of the GNU General Public 
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include <stdio.h>
#include <stdlib.h>
#include <new>

#include "positionDB.H"
#include "existDB.H"
#include "bri++.H"

//#define ERROR_CHECK_COUNTING
//#define ERROR_CHECK_COUNTING_ENCODING
//#define ERROR_CHECK_EMPTY_BUCKETS

//  This tests Chunlin Xiao's discovered bug -- if there are a small
//  number of unique mers, compared to distinct mers (2 * #unique_mers
//  < #distinct_mers, we would overflow the position pointer in
//  buckets.  This enables a check that it doesn't occur.
//
//#define TEST_NASTY_BUGS

#define MSG_OUTPUT  stderr


positionDB::positionDB(char const    *filename,
                       bool           loadData) {
  bool fail = false;

  if (loadState(filename, true, loadData) == false) {
    fprintf(stderr, "positionDB::positionDB()-- Tried to read state from '%s', but failed.\n", filename);
    exit(1);
  }

#if 0
  if (_merSizeInBases != merSize) {
    fprintf(stderr, "positionDB::positionDB()-- Read state from '%s', but got different mer sizes\n", filename);
    fprintf(stderr, "positionDB::positionDB()-- Got "u32bitFMT", expected "u32bitFMT"\n", _merSizeInBases, merSize);
    fail = true;
  }
  if (_merSkipInBases != merSkip) {
    fprintf(stderr, "positionDB::positionDB()-- Read state from '%s', but got different mer skips\n", filename);
    fprintf(stderr, "positionDB::positionDB()-- Got "u32bitFMT", expected "u32bitFMT"\n", _merSkipInBases, merSkip);
    fail = true;
  }
  if (_tableSizeInBits != tblBits) {
    fprintf(stderr, "positionDB::positionDB()-- Read state from '%s', but got different table sizes\n", filename);
    fprintf(stderr, "positionDB::positionDB()-- Got "u32bitFMT", expected "u32bitFMT"\n", _tableSizeInBits, tblBits);
    fail = true;
  }
#endif

  if (fail)
    exit(1);
}


positionDB::positionDB(merStream   *MS,
                       u32bit       merSize,
                       u32bit       merSkip,
                       u32bit       tblBits,
                       existDB     *mask,
                       existDB     *only,
                       bool         beVerbose) {

  _bucketSizes           = 0L;
  _countingBuckets       = 0L;
  _hashTable             = 0L;
  _buckets               = 0L;
  _positions             = 0L;

  _merSizeInBases        = merSize;
  _merSizeInBits         = 2 * _merSizeInBases;
  _merSkipInBases        = merSkip;
  _tableSizeInBits       = tblBits;
  _tableSizeInEntries    = u64bitONE << _tableSizeInBits;
  _hashWidth             = u32bitZERO;
  _hashMask              = u64bitMASK(_tableSizeInBits);
  _chckWidth             = _merSizeInBits - _tableSizeInBits;
  _chckMask              = u64bitMASK(_chckWidth);
  _posnWidth             = u64bitZERO;
  _posnMask              = u64bitZERO;

  _wCnt                  = 0;
  _wFin                  = 0;

  _shift1                = _merSizeInBits - _tableSizeInBits;
  _shift2                = _shift1 / 2;
  _mask1                 = u64bitMASK(_tableSizeInBits);
  _mask2                 = u64bitMASK(_shift1);

  _numberOfMers          = u64bitZERO;
  _numberOfPositions     = u64bitZERO;
  _numberOfDistinct      = u64bitZERO;
  _numberOfUnique        = u64bitZERO;
  _numberOfEntries       = u64bitZERO;
  _maximumEntries        = u64bitZERO;

  _sortedListMax         = u32bitZERO;
  _sortedListLen         = u32bitZERO;
  _sortedList            = 0L;


  if (MS == 0L) {
    fprintf(stderr, "positionDB()-- ERROR: No merStream?  Nothing to build a table with!\n");
    exit(1);
  }

  if (MS->rewind() == false) {
    fprintf(stderr, "positionDB::positionDB(): Failed initial rewind of merStream!\n");
    fprintf(stderr, "positionDB::positionDB(): Your merStream doesn't support rewind?!!\n");
    exit(1);
  }



  ////////////////////////////////////////////////////////////////////////////////
  //
  //  1)  Count bucket sizes
  //

  //  We'll later want to reuse the _bucketSizes space for storing
  //  the _hashTable.  To make it somewhat safe, we allocate the
  //  space as u64bit, then cast it to be u32bit.
  //
  //  bktAllocIsJunk tells us if we should release this memory (if
  //  we need to allocate separate space for the _hashTable).
  //
  //  The _bucketSizes is offset by one from bktAlloc so that
  //  we don't overwrite _bucketSizes when we are constructing
  //  _hashTable.
  //
  u64bit *bktAlloc;
  try {
    bktAlloc = new u64bit [_tableSizeInEntries / 2 + 2];
  } catch (std::bad_alloc) {
    fprintf(stderr, "hitMatrix::filter()-- caught std::bad_alloc in %s at line %d\n", __FILE__, __LINE__);
    fprintf(stderr, "hitMatrix::filter()-- bktAlloc = new u64bit ["u64bitFMT"]\n", _tableSizeInEntries / 2 + 2);
    exit(1);
  }
  bool     bktAllocIsJunk = false;

  for (u64bit i=_tableSizeInEntries / 2 + 2; i--; )
    bktAlloc[i] = 0;

  _bucketSizes = (u32bit *)(bktAlloc + 1);

#ifdef ERROR_CHECK_COUNTING
  fprintf(stdout, "ERROR_CHECK_COUNTING is defined.\n");
  u32bit *_errbucketSizes = new u32bit [_tableSizeInEntries + 2];
  for (u64bit i=_tableSizeInEntries + 2; i--; )
    _errbucketSizes[i] = 0;
#endif

  if (beVerbose)
    fprintf(MSG_OUTPUT, "    Allocated bucket size counting space with total size "u64bitFMT" KB\n", _tableSizeInEntries >> 8);


  speedCounter  *C = new speedCounter("    %7.2f Mmers -- %5.2f Mmers/second\r", 1000000.0, 0x1fffff, beVerbose);

  //  Two choices here
  //
  //  1)  No masking or onlying is done.  Stream the mers and just
  //      count the positions.  This is the original behavior.
  //
  //  2)  Masking or onlying is done.  Open the output stream file,
  //      stream the mers by, checking for mask/only of both
  //      forward and reverse mers.  If either is found, push
  //      the (forward) mer and position onto the stream.
  //      close the output stream.
  //
  //      Save the mer if it doesn't exist in the mask (both f and r),
  //      or does exist in the only (either f or r), add it.
  //
  //      The input databases for mask and only are (currently) made
  //      using canonical mers.  We halve the number of exists() by
  //      also using canonical mers here.
  //

  if (MS->rewind() == false) {
    fprintf(stderr, "positionDB::positionDB(): Failed second rewind of merStream!\n");
    exit(1);
  }

  if (mask) {
    while (MS->nextMer(_merSkipInBases)) {
      u64bit  canonicalmer = MS->theFMer();
      if (canonicalmer > MS->theRMer())
        canonicalmer = MS->theRMer();

      if (!mask->exists(canonicalmer)) {
        _bucketSizes[ HASH(MS->theFMer()) ]++;
        _numberOfMers++;
        _numberOfPositions = MS->thePositionInStream();
      }

      C->tick();
    }
  } else if (only) {
    while (MS->nextMer(_merSkipInBases)) {
      u64bit  canonicalmer = MS->theFMer();
      if (canonicalmer > MS->theRMer())
        canonicalmer = MS->theRMer();

      if (only->exists(canonicalmer)) {
        _bucketSizes[ HASH(MS->theFMer()) ]++;
        _numberOfMers++;
        _numberOfPositions = MS->thePositionInStream();
      }

      C->tick();
    }
  } else {
    while (MS->nextMer(_merSkipInBases)) {
      _bucketSizes[ HASH(MS->theFMer()) ]++;

#ifdef ERROR_CHECK_COUNTING
      _errbucketSizes[ HASH(MS->theFMer()) ]++;
#endif

      _numberOfMers++;
      _numberOfPositions = MS->thePositionInStream();
      C->tick();
    }
  }

  delete C;
  C = 0L;

  if (beVerbose)
    fprintf(MSG_OUTPUT, "    Found "u64bitFMT" mers (max position = "u64bitFMT")\n", _numberOfMers, _numberOfPositions);

  //  This is _numberOfMers+1 because we need to store the first
  //  position after the last mer.  That is, if there are two mers, we
  //  will store that the first mer is at position 0, the second mer
  //  is at position 1, and the end of the second mer is at position 2.
  //
  _hashWidth = 1;
  while ((_numberOfMers+1) > (u64bitONE << _hashWidth))
    _hashWidth++;

  _posnWidth = 1;
  while ((_numberOfPositions+1) > (u64bitONE << _posnWidth))
    _posnWidth++;
  _posnMask = u64bitMASK(_posnWidth);



  ///////////////////////////////////////////////////////////////////////////////
  //
  //  2)  Allocate buckets and make bucketSizes be a pointer into them
  //
  _wCnt          = _chckWidth + _posnWidth + 1;

  u64bit   bucketsSpace  = (_numberOfMers+1) * _wCnt / 64 + 1;
  u32bit   endPosition   = 0;

  if (_wCnt > 64) {
    fprintf(stderr, "ERROR: Data sizes to big: wCnt="u32bitFMT", should be <= 64.\n", _wCnt);
    fprintf(stderr, "       Reduce mersize, number of mers, or both.\n");
    exit(1);
  }


  if (beVerbose)
    fprintf(MSG_OUTPUT, "    Allocated "u64bitFMT"KB for buckets ("u64bitFMT" 64-bit words)\n", bucketsSpace >> 7, bucketsSpace);
  try {
    _countingBuckets = new u64bit [bucketsSpace];
  } catch (std::bad_alloc) {
    fprintf(stderr, "hitMatrix::filter()-- caught std::bad_alloc in %s at line %d\n", __FILE__, __LINE__);
    fprintf(stderr, "hitMatrix::filter()-- _countingBuckets = new u64bit ["u64bitFMT"]\n", bucketsSpace);
    exit(1);
  }

  for (u64bit i=bucketsSpace; i--; )
    _countingBuckets[i] = ~u64bitZERO;

  for (u32bit i=0; i<_tableSizeInEntries; i++) {
    endPosition     += _bucketSizes[i];
    _bucketSizes[i]  = endPosition;
  }
  _bucketSizes[_tableSizeInEntries] = endPosition;

#ifdef ERROR_CHECK_COUNTING
  fprintf(stdout, "ERROR_CHECK_COUNTING:  endposition = "u32bitFMT"\n", endPosition);
  fprintf(stdout, "ERROR_CHECK_COUNTING:  nummers     = "u64bitFMT"\n", _numberOfMers);
  if (endPosition != _numberOfMers)
    fprintf(stdout, "ERROR_CHECK_COUNTING: BUCKETSIZE COUNTING PROBLEM -- endPos != numMers\n");
#endif


  ////////////////////////////////////////////////////////////////////////////////
  //
  //  3)  Build list of mers with positions
  //
  if (beVerbose)
    fprintf(MSG_OUTPUT, "    Building lists with positions.\n");

  C = new speedCounter("    %7.2f Mmers -- %5.2f Mmers/second\r", 1000000.0, 0x1fffff, beVerbose);

#ifdef ERROR_CHECK_COUNTING_ENCODING
  fprintf(stdout, "ERROR_CHECK_COUNTING_ENCODING is defined!\n");
#endif


  if (MS->rewind() == false) {
    fprintf(stderr, "positionDB::positionDB(): Failed third rewind of merStream!\n");
    exit(1);
  }

  //
  //  Grrr.  We used to have the mers filtered already, but no more.
  //  So, we need to do the ugly masking again.  Debugging checks are
  //  only enabled for _unmasked_ builds (seatac is currently the only
  //  client that uses masking)
  //

  if (mask) {

    while (MS->nextMer(_merSkipInBases)) {
      u64bit  canonicalmer = MS->theFMer();
      if (canonicalmer > MS->theRMer())
        canonicalmer = MS->theRMer();

      if (!mask->exists(canonicalmer)) {
        u64bit h = HASH(MS->theFMer());
        _bucketSizes[h]--;
        setDecodedValue(_countingBuckets, (u64bit)_bucketSizes[h] * (u64bit)_wCnt, _wCnt,
                        (CHECK(MS->theFMer()) << _posnWidth) | (MS->thePositionInStream() & _posnMask));
      }

      C->tick();
    }


  } else if (only) {

    while (MS->nextMer(_merSkipInBases)) {
      u64bit  canonicalmer = MS->theFMer();
      if (canonicalmer > MS->theRMer())
        canonicalmer = MS->theRMer();

      if (only->exists(canonicalmer)) {
        u64bit h = HASH(MS->theFMer());
        _bucketSizes[h]--;
        setDecodedValue(_countingBuckets, (u64bit)_bucketSizes[h] * (u64bit)_wCnt, _wCnt,
                        (CHECK(MS->theFMer()) << _posnWidth) | (MS->thePositionInStream() & _posnMask));
      }

      C->tick();
    }

  } else {
    while (MS->nextMer(_merSkipInBases)) {
      u64bit h = HASH(MS->theFMer());

#ifdef ERROR_CHECK_COUNTING
      if (_bucketSizes[h] == 0) {
        fprintf(stderr, "ERROR_CHECK_COUNTING: Bucket "u64bitFMT" ran out of things!  '%s'\n", h, MS->theFMerString());
        fprintf(stderr, "ERROR_CHECK_COUNTING: Stream is at "u64bitFMT"\n", MS->thePositionInStream());
      }
#endif

      _bucketSizes[h]--;

#ifdef ERROR_CHECK_COUNTING
      _errbucketSizes[h]--;
#endif


#ifdef ERROR_CHECK_EMPTY_BUCKETS
      if ((~getDecodedValue(_countingBuckets, (u64bit)_bucketSizes[h] * (u64bit)_wCnt, _wCnt)) & u64bitMASK(_wCnt))
        fprintf(stdout, "ERROR_CHECK_EMPTY_BUCKETS: countingBucket not empty!  pos=%lu\n", _bucketSizes[h] * _wCnt);
#endif

      setDecodedValue(_countingBuckets, (u64bit)_bucketSizes[h] * (u64bit)_wCnt, _wCnt,
                      (CHECK(MS->theFMer()) << _posnWidth) | (MS->thePositionInStream() & _posnMask));


#ifdef ERROR_CHECK_COUNTING_ENCODING
      if ((MS->thePositionInStream() & _posnMask) != MS->thePositionInStream())
        fprintf(stdout, "ERROR_CHECK_COUNTING_ENCODING error:  POSNMASK "u64bitHEX" invalid!  Wanted "u64bitHEX" got "u64bitHEX"\n",
                _posnMask,
                MS->thePositionInStream(),
                MS->thePositionInStream() & _posnMask);
#endif

#ifdef ERROR_CHECK_COUNTING_ENCODING
      u64bit v = getDecodedValue(_countingBuckets, (u64bit)_bucketSizes[h] * (u64bit)_wCnt, _wCnt);

      //  This test is only valid if we have an extra bit at the start --
      //  if we are planning on reusing the counting space for buckets.
      //
      if ((0 != (v >> (_wCnt - 1))))
        fprintf(stdout, "ERROR_CHECK_COUNTING_ENCODING error: HBIT is set!      Wanted "u64bitHEX" got "u64bitHEX"\n",
                (CHECK(MS->theFMer()) << _posnWidth) | (MS->thePositionInStream() & _posnMask), v);
      if (CHECK(MS->theFMer()) != ((v >> _posnWidth) & _chckMask))
        fprintf(stdout, "ERROR_CHECK_COUNTING_ENCODING error:  CHCK corrupted!  Wanted "u64bitHEX" got "u64bitHEX"\n",
                (CHECK(MS->theFMer()) << _posnWidth) | (MS->thePositionInStream() & _posnMask), v);
      if (MS->thePositionInStream() != (v & _posnMask))
        fprintf(stdout, "ERROR_CHECK_COUNTING_ENCODING error:  POSN corrupted!  Wanted "u64bitHEX" got "u64bitHEX"\n",
                (CHECK(MS->theFMer()) << _posnWidth) | (MS->thePositionInStream() & _posnMask), v);
#endif

      C->tick();
    }
  }

  delete C;
  C = 0L;

#ifdef ERROR_CHECK_COUNTING
  fprintf(stdout, "Checking for unfilled buckets\n");
  for (u32bit i=0; i<_tableSizeInEntries; i++)
    if (_errbucketSizes[i] != 0)
      fprintf(stdout, "ERROR_CHECK_COUNTING: Bucket "u32bitFMT" wasn't filled fully?  "u32bitFMT" left over.\n", i, _errbucketSizes[i]);
  fprintf(stdout, "ERROR_CHECK_COUNTING\n");
#endif


  ////////////////////////////////////////////////////////////////////////////////
  //
  //  4)  Sort each bucket -- count:
  //        1) number of distinct mers
  //        2) number of unique mers
  //        3) number of entries in position table ( sum mercount+1 for all mercounts > 1)
  //      also need to repack the sorted things
  //
  if (beVerbose)
    fprintf(MSG_OUTPUT, "    Sorting and repacking buckets.\n");

  for (u32bit i=0; i<_tableSizeInEntries; i++)
    sortAndRepackBucket(i);

  if (beVerbose)
    fprintf(MSG_OUTPUT,
            "    Found "u64bitFMTW(12)" total mers\n"
            "    Found "u64bitFMTW(12)" distinct mers\n"
            "    Found "u64bitFMTW(12)" unique mers\n"
            "    Need "u64bitFMT" non-unique position list entries ("u64bitFMT" maximum count)\n",
            _numberOfMers, _numberOfDistinct, _numberOfUnique, _numberOfEntries, _maximumEntries);



  ////////////////////////////////////////////////////////////////////////////////
  //
  //  Compute the size of the final bucket position entry.  It's
  //  either a position into the sequence, or a pointer into a list of
  //  positions.  In rare cases, the pointer is larger than the
  //  sequence position, and we need to do extra work.
  //
  _posnPtrWidth = 1;
  _posnPtrMask  = 0;
  while ((_numberOfEntries+1) > (u64bitONE << _posnPtrWidth))
    _posnPtrWidth++;

  //  The width of position pointers (in buckets) is the max of
  //  _posnWidth (a pointer to the sequence position) and
  //  _posnPtrWidth (a pointer to an entry in the positions table).
  //
  if (_posnPtrWidth < _posnWidth)
    _posnPtrWidth = _posnWidth;

  _posnPtrMask = u64bitMASK(_posnPtrWidth);
  
  _wFin = _chckWidth + _posnPtrWidth + 1;


  ////////////////////////////////////////////////////////////////////////////////
  //
  //  5)  Allocate: real hash table, buckets and position table.
  //
  u64bit  hs = _tableSizeInEntries * _hashWidth / 64 + 1;
  u64bit  bs = _numberOfDistinct   * _wFin      / 64 + 1;
  u64bit  ps = _numberOfEntries    * _posnWidth / 64 + 1;

  if (_hashWidth < 32) {
    if (beVerbose)
      fprintf(MSG_OUTPUT, "    Reusing bucket counting space for hash table.\n");

    _hashTable     = bktAlloc;
    bktAllocIsJunk = false;
  } else {
    if (beVerbose)
      fprintf(MSG_OUTPUT, "    Allocated "u64bitFMTW(10)"KB for hash table ("u64bitFMT" 64-bit words)\n",
              hs >> 7, hs);

    try {
      _hashTable     = new u64bit [hs];
    } catch (std::bad_alloc) {
      fprintf(stderr, "hitMatrix::filter()-- caught std::bad_alloc in %s at line %d\n", __FILE__, __LINE__);
      fprintf(stderr, "hitMatrix::filter()-- _hashTable = new u64bit ["u64bitFMT"]\n", hs);
      exit(1);
    }
    bktAllocIsJunk = true;
  }


  //  If we have enough space to reuse the counting space, reuse it.
  //  Else, allocate more space.
  //
  //  We need to ensure that there are enough bits and that the size
  //  of a bucket didn't increase.  If the bucket size did increase,
  //  and we see more unique buckets than total mers (up to some
  //  point) we overwrite data.
  //
  //  Recall that bucketSpace ~= numberOfMers * wCnt
  //
  if ((bs < bucketsSpace) && (_wFin <= _wCnt)) {
    if (beVerbose)
      fprintf(MSG_OUTPUT, "    Reusing bucket space; Have: "u64bitFMT"  Need: "u64bitFMT" (64-bit words)\n", bucketsSpace, bs);

    _buckets = _countingBuckets;

    bs = bucketsSpace; // for output at the end
  } else {
    if (beVerbose)
      fprintf(MSG_OUTPUT, "    Allocated "u64bitFMTW(10)"KB for buckets    ("u64bitFMT" 64-bit words)\n", bs >> 7, bs);

    try {
      _buckets   = new u64bit [bs];
    } catch (std::bad_alloc) {
      fprintf(stderr, "hitMatrix::filter()-- caught std::bad_alloc in %s at line %d\n", __FILE__, __LINE__);
      fprintf(stderr, "hitMatrix::filter()-- _buckets = new u64bit ["u64bitFMT"]\n", bs);
      exit(1);
    }
  }

  if (beVerbose)
    fprintf(MSG_OUTPUT, "    Allocated "u64bitFMTW(10)"KB for positions  ("u64bitFMT" 64-bit words)\n", ps >> 7, ps);
  try {
    _positions = new u64bit [ps];
  } catch (std::bad_alloc) {
    fprintf(stderr, "hitMatrix::filter()-- caught std::bad_alloc in %s at line %d\n", __FILE__, __LINE__);
    fprintf(stderr, "hitMatrix::filter()-- _positions = new u64bit ["u64bitFMT"\n", ps);
    exit(1);
  }

  ////////////////////////////////////////////////////////////////////////////////
  //
  //  6)  Transfer from the sorted buckets to the hash table.
  //
  if (beVerbose)
    fprintf(MSG_OUTPUT, "    Transferring to final structure.\n");

  u64bit   bucketStartPosition = 0;
  u64bit   checkMask = ~_posnMask;

  //  Current positions and bit positions in the buckets and position list.
  //
  u64bit  currentBbit = u64bitZERO;  //  Bit position into bucket
  u64bit  currentPbit = u64bitZERO;  //  Bit position into positions
  u64bit  currentPpos = u64bitZERO;  //  Value position into positions

#ifdef TEST_NASTY_BUGS
  //  Save the position array pointer of each bucket for debugging.
  //
  u64bit  currentBpos = u64bitZERO;  //  Value position into bucket
  u32bit *posPtrCheck = new u32bit [65826038];
#endif


  //  We need b outside the loop!
  //
  u32bit  b;
  for (b=0; b<_tableSizeInEntries; b++) {

    //  Set the start of the bucket -- we took pains to ensure that
    //  we don't overwrite _bucketSizes[b], if we are reusing that
    //  space for _hashTable.
    //
    setDecodedValue(_hashTable, (u64bit)b * (u64bit)_hashWidth, _hashWidth, bucketStartPosition);

    //  Get the number of mers in the counting bucket.  The error checking
    //  and sizing of _sortedList was already done in the sort.
    //
    u64bit st = _bucketSizes[b];
    u64bit ed = _bucketSizes[b+1];
    u64bit le = ed - st;

    //  NOTE:  If all the mers are unique, _sortedList will not be allocated
    //  in sortAndRepackBucket().  In that case, we need to allocate some space.
    //
    if (_sortedList == 0L) {
      _sortedListMax = 1024;
      _sortedListLen = u32bitZERO;
      _sortedList    = new heapbit [_sortedListMax];
    }

    //  Unpack the check values
    //
    for (u64bit i=st, J=st * _wCnt; i<ed; i++, J += _wCnt)
      _sortedList[i-st] = getDecodedValue(_countingBuckets, J, _wCnt);

    //
    //  Walk through the counting bucket, adding things to the real
    //  bucket as we see them.  Mers with more than one position are
    //  inserted into the bucket, and the positions inserted into the
    //  position list.
    //

    //  start and end locations of the mer.  For mers with only
    //  one occurrance (unique mers), stM+1 == edM.
    //
    u64bit  stM = u64bitZERO;
    u64bit  edM = u64bitZERO;
    u64bit  v;

    while (stM < le) {

      //  Move to the next mer.
      //
      edM++;

      //  Keep moving while the two mers are the same.
      //
      while ((edM < le) &&
             ((_sortedList[stM] & checkMask) == (_sortedList[edM] & checkMask)))
        edM++;

      //  edM is now the mer after the last.  Write all mers from stM up to edM
      //  to the final structure.  If there is one mer, put it in the bucket.
      //  If not, put a pointer to the position array there.
      //
      if (stM+1 == edM) {
#ifdef TEST_NASTY_BUGS
        posPtrCheck[currentBpos++] = _sortedList[stM] & _posnMask;
#endif

        //  Rearrange the check and position, insert a bit telling us that
        //  this is a position and not an offset.
        //
        v  = u64bitONE << (_wFin - 1);
        v |= (_sortedList[stM] & _posnMask) << _chckWidth;
        v |= (_sortedList[stM] & checkMask) >> _posnWidth;

        setDecodedValue(_buckets, currentBbit, _wFin, v);
        currentBbit += _wFin;
        bucketStartPosition++;
      } else {
#ifdef TEST_NASTY_BUGS
        posPtrCheck[currentBpos++] = currentPpos;
 #endif

        //  Create a pointer to the position list.
        //
        v  = u64bitZERO;
        v |= (currentPpos      & _posnPtrMask) << _chckWidth;
        v |= (_sortedList[stM] & checkMask)    >> _posnWidth;

        setDecodedValue(_buckets, currentBbit, _wFin, v);
        currentBbit += _wFin;
        bucketStartPosition++;

        //  Store the positions.
        //
        setDecodedValue(_positions, currentPbit, _posnWidth, edM - stM);
        currentPbit += _posnWidth;
        currentPpos++;

        //  The positions are in the proper place in _sortedList, and setDecodedValue
        //  masks out the extra crap, so no temporary needed.
        //
        for (; stM < edM; stM++) {
          setDecodedValue(_positions, currentPbit, _posnWidth, _sortedList[stM]);
          currentPbit += _posnWidth;
          currentPpos++;
        }
      }

      //  All done with this mer.
      //
      stM = edM;
    }
  }

  //  Set the end of the last bucket
  //
  setDecodedValue(_hashTable, b * _hashWidth, _hashWidth, bucketStartPosition);

  if (beVerbose) {
    fprintf(MSG_OUTPUT, "    Avail: Bucket "u64bitFMTW(12)"    Position "u64bitFMTW(12)" (64-bit words)\n", bs, ps);
    fprintf(MSG_OUTPUT, "    Used:  Bucket "u64bitFMTW(12)"    Position "u64bitFMTW(12)" (64-bit words)\n", currentBbit / 64, currentPbit / 64);
  }


#ifdef TEST_NASTY_BUGS
  //  Unpack the bucket positions and check.  Report the first one that is broken.
  //
  for(u64bit b=0; b<currentBpos; b++) {
    u64bit c = (getDecodedValue(_buckets, b * _wFin, _wFin) >> _chckWidth) & _posnPtrMask;
    if (posPtrCheck[b] != c) {
      fprintf(stderr, "Bucket %u (at bitpos %lu) failed position check (wanted %u got %u)\n",
              (u32bit)b,
              b * _wFin,
              (u32bit)posPtrCheck[b],
              (u32bit)c);
    }
  }

  delete [] posPtrCheck;
#endif


  //  Free the counting buckets is we aren't using the space for
  //  something else.
  //
  if (_buckets != _countingBuckets)
    delete [] _countingBuckets;

  //  In theory, we could move these to be immediately after the data
  //  is useless.
  //
  _bucketSizes     = 0L;
  _countingBuckets = 0L;

  delete [] _sortedList;

  _sortedListMax = 0;
  _sortedListLen = 0;
  _sortedList    = 0L;

  if (bktAllocIsJunk)
    delete [] bktAlloc;
}

positionDB::~positionDB() {
  delete [] _hashTable;
  delete [] _buckets;
  delete [] _positions;
}

