// This file is part of ESTmapper.
// Copyright (c) 2001, 2002, 2003, 2004 Applera Corporation
// Authors: Brian Walenz and Liliana Florea
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received (LICENSE.txt) a copy of the GNU General Public 
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include "hitReader.H"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>


static
int
hitCompare(const void *a, const void *b) {
  const hit_s  *A = (const hit_s *)a;
  const hit_s  *B = (const hit_s *)b;

  if (A->coverage > B->coverage)
    return(-1);
  else
    return(A->coverage < B->coverage);
}







hitReader::hitReader(int m) {
  _filesMax   = m;
  _filesLen   = 0;
  _files      = new hitFile_s [_filesMax];

  _listLen    = 0;
  _listMax    = 1024 * 1024;
  _list       = new hit_s [_listMax];

  _iid        = u32bitZERO;
  _bestScore  = 0.0;
  _worstScore = 1.0;
}


hitReader::~hitReader() {
  for (u32bit i=0; i<_filesLen; i++)
    fclose(_files[i].file);

  delete [] _files;
  delete [] _list;
}

void
hitReader::addInputFile(char *filename) {
  errno = 0;

  _files[_filesLen].stillMore = true;

  if (strcmp(filename, "-") == 0) {
    _files[_filesLen].file = stdin;
  } else {
    _files[_filesLen].file = fopen(filename, "r");
  }

  if (_files[_filesLen].file == 0L) {
    fprintf(stderr, "hitReader::addInputFile()-- ERROR: couldn't open '%s' for reading.\n", filename);
    fprintf(stderr, "hitReader::addInputFile()-- %s\n", strerror(errno));
    exit(1);
  }

  //  Binary or ASCII input?
  //
  char x = (char)fgetc(_files[_filesLen].file);
  ungetc(x, _files[_filesLen].file);

  _files[_filesLen].isBINARY = (x != '-');

  //  Load the first hit
  loadHit(_files+_filesLen);

  _filesLen++;
}


void
hitReader::loadHit(hitFile_s *HF) {
  if (HF->isBINARY) {
    ahit_readBinary(&HF->a, HF->file);
  } else {
    fgets(HF->b, 1024, HF->file);
    ahit_parseString(&HF->a, HF->b);
  }


  if (feof(HF->file))
    HF->stillMore = false;
};


bool
hitReader::loadHits(void) {

  _listLen    = 0;
  _iid        = u32bitZERO;
  _bestScore  = 0.0;
  _worstScore = 1.0;

  //  See if there are more hits to process.
  //
  bool  keepGoing = false;
  for (u32bit i=0; i<_filesLen; i++)
    keepGoing |= _files[i].stillMore;

  if (keepGoing == false)
    return(false);

  //  Find the lowest ESTid
  //
  _iid = 1 << 30;
  for (u32bit i=0; i<_filesLen; i++)
    if ((_files[i].stillMore) && (_iid > _files[i].a._qsIdx))
      _iid = _files[i].a._qsIdx;


  //  For each file, load the next hit if it's the est
  //  we're looking at
  //
  for (u32bit i=0; i<_filesLen; i++) {
    while ((_files[i].stillMore) && (_files[i].a._qsIdx == _iid)) {
      grow_List();

      memcpy(&_list[_listLen].a, &_files[i].a, sizeof(aHit));

      _list[_listLen].coverage     = (double)_files[i].a._covered / (double)_files[i].a._numMers;
      _list[_listLen].multiplicity = (double)_files[i].a._matched / (double)_files[i].a._covered;

      //  aHit->_covered is in bases, but aHit->_numMers is the
      //  number of mers.  Possible for coverage to be > 1.0.
      //
      if (_list[_listLen].coverage > 1.0)
        _list[_listLen].coverage = 1.0;

      if (_list[_listLen].coverage > _bestScore)
        _bestScore = _list[_listLen].coverage;

      if (_list[_listLen].coverage < _worstScore)
        _worstScore = _list[_listLen].coverage;

#ifdef WITH_ANSWERS
      //  Look for the answer string.  If not found, set to zero.
      //
      _list[_listLen].mappedIdentity = 0;
      _list[_listLen].mappedCoverage = 0;

      for (int p=0; _files[i].b[p]; p++) {
        if ((_files[i].b[p] == 'Y') || (_files[i].b[p] == 'N')) {
          char *c = _files[i].b+p+1;
          _list[_listLen].mappedIdentity = (u32bit)strtoul(c, &c, 10);
          _list[_listLen].mappedCoverage = (u32bit)strtoul(c, &c, 10);
        }
      }
#endif

      _listLen++;

      loadHit(_files+i);
    }
  }

  return(true);
}



void
hitReader::sortByCoverage(void) {
  qsort(_list, _listLen, sizeof(hit_s), hitCompare);
};
