// This file is part of ESTmapper.
// Copyright (c) 2001, 2002, 2003, 2004 Applera Corporation
// Authors: Brian Walenz and Liliana Florea
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received (LICENSE.txt) a copy of the GNU General Public 
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include "thr.H"


void
doSearch(searcherState       *state,
         FastASequenceInCore *seq,
         u32bit               idx,
         bool                 rc,
         aHit               *&theHits,
         u32bit              &theHitsLen,
         u32bit              &theHitsMax,
         logMsg              *theLog) {
  encodedQuery  *query  = 0L;
  hitMatrix     *matrix = 0L;
  u32bit         qMers  = 0;
  double         startTime  = 0.0;


  ///////////////////////////////////////
  //
  //  Build and mask the query
  //
  startTime = getTime();
  query = new encodedQuery(seq->sequence(), seq->sequenceLength(), config._merSize, rc);

  startTime = getTime();
  if (maskDB)
    for (u32bit qi=0; qi<query->numberOfMers(); qi++)
      if ((query->getSkip(qi) == false) &&
          (maskDB->exists(query->getMer(qi))))
        query->setSkip(qi);

  if (onlyDB)
    for (u32bit qi=0; qi<query->numberOfMers(); qi++)
      if ((query->getSkip(qi) == false) &&
          (!onlyDB->exists(query->getMer(qi))))
        query->setSkip(qi);

#ifdef USEEXACTSIZE
  merCovering   *IL = new merCovering(config._merSize);

  for (u32bit qi=0; qi<query->numberOfMers(); qi++) {
    if (query->getSkip(qi) == false)
      IL->addMer(qi);
  }

  qMers = IL->sumLengths();
  delete IL;
#else
  qMers = query->numberOfValidMers();
#endif
  state->maskTime += getTime() - startTime;




#ifdef SAVE_HITS_TO_FILES
  {
    char  filename[1024];
    if (rc)
      sprintf(filename, "hitsDump-%08d-r.1.raw", (int)seq->getIID());
    else
      sprintf(filename, "hitsDump-%08d-f.1.raw", (int)seq->getIID());
    errno = 0;
    state->hitsDump = fopen(filename, "w");
    if (errno) {
      fprintf(stderr, "Failed to open '%s' for hitsDump!\n", filename);
      exit(1);
    }
  }
#endif




  ///////////////////////////////////////
  //
  //  Get the hits
  //
  startTime = getTime();
  matrix = new hitMatrix(seq->sequenceLength(), qMers, idx);
  for (u32bit qi=0; qi<query->numberOfMers(); qi++) {
    if ((query->getSkip(qi) == false) &&
        (positions->get(query->getMer(qi), state->posn, state->posnMax, state->posnLen))) {
#ifdef SAVE_HITS_TO_FILES
      {
        for (u32bit x=0; x<state->posnLen; x++)
          fprintf(state->hitsDump, u32bitFMT" "u64bitFMT"\n", qi, state->posn[x]);
      }
#endif
      matrix->addHits(qi, state->posn, state->posnLen);
    }
  }
  state->searchTime += getTime() - startTime;




  ///////////////////////////////////////
  //
  //  Chain the hits
  //
  startTime = getTime();
  matrix->filter(rc ? 'r' : 'f', config._minHitCoverage, config._minHitLength,
                 theHits, theHitsLen, theHitsMax);
  state->chainTime += getTime() - startTime;



#ifdef SAVE_HITS_TO_FILES
  {
    fclose(state->hitsDump);

    char  filename[1024];
    if (rc)
      sprintf(filename, "hitsDump-%08d-r.2.refine", (int)seq->getIID());
    else
      sprintf(filename, "hitsDump-%08d-f.2.refine", (int)seq->getIID());
    errno = 0;
    state->hitsDump = fopen(filename, "w");
    if (errno) {
      fprintf(stderr, "Failed to open '%s' for hitsDump!\n", filename);
    }
  }
#endif


  ////////////////////////////////////////
  //
  //  Refine the hits -- if any hit looks like it contains a repeat,
  //  rebuild it using an adaptive mask threshold.
  //
  //  We work backwards because we add on new hits to the end of our
  //  list.
  //
  for (u32bit h=theHitsLen; h--; ) {

    if (((theHits[h]._status & AHIT_DIRECTION_MASK) == !rc) && 
        (theHits[h]._matched > 2 * theHits[h]._numMers)) {

#ifdef SHOW_HIT_DISCARDING
      theLog->add("Seq "u32bitFMT" Hit "u32bitFMT" (%c) has "u32bitFMT" matched, but only "u32bitFMT" mers.\n",
                 seq->getIID(), h, rc ? 'r' : 'f', theHits[h]._matched, theHits[h]._numMers);
#endif

      //  Grab the genomic sequence
      //
      FastASequenceInCore  *GENseq = cache->getSequence(theHits[h]._dsIdx);
      u32bit                GENlo  = theHits[h]._dsLo;
      u32bit                GENhi  = theHits[h]._dsHi;

      //  Construct a merstream for the region
      //
      merStream            *MS     = new merStream(config._merSize, GENseq->sequence() + GENlo, GENhi - GENlo);

      //  Build a positionDB of the region.  This gets us both
      //  positions and counts.  Then fill out another hitMatrix,
      //  using about 2*length mers.
      //
      positionDB           *PS     = new positionDB(MS,
                                                    config._merSize,
                                                    0,
                                                    20,
                                                    0L, 0L, false);

      hitMatrix            *HM = new hitMatrix(seq->sequenceLength(), qMers, idx);

      //  We find the number of hits we would get if we use a
      //  countLimit of i.
      //
#define COUNT_MAX   256

      u32bit numHitsAtCount[COUNT_MAX] = { 0 };
      u32bit countLimit                = 0;

      u32bit numMers = 0;
      u32bit numHits = 0;
      u32bit minNum  = ~u32bitZERO;
      u32bit maxNum  = 0;

      for (u32bit qi=0; qi<query->numberOfMers(); qi++) {
        if ((query->getSkip(qi) == false) &&
            (PS->get(query->getMer(qi), state->posn, state->posnMax, state->posnLen))) {

          if (state->posnLen-1 < COUNT_MAX)
            numHitsAtCount[state->posnLen-1]++;

          //  Collect some statistics

          numMers++;
          numHits += state->posnLen;

          if (minNum > state->posnLen)  minNum = state->posnLen;
          if (maxNum < state->posnLen)  maxNum = state->posnLen;
        }
      }

      numHitsAtCount[0] = 0;

      for (u32bit qi=1; qi<COUNT_MAX; qi++) {
        numHitsAtCount[qi] = numHitsAtCount[qi-1] + qi * numHitsAtCount[qi];

        if (numHitsAtCount[qi] < 50000)
          countLimit = qi;
      }

#ifdef SHOW_HIT_DISCARDING
      theLog->add(" -- found "u32bitFMT" hits in "u32bitFMT" mers, min="u32bitFMT" max="u32bitFMT" avg=%.5f hits/mer.\n",
                 numHits, numMers, minNum, maxNum, (double)numHits / (double)numMers);
      theLog->add(" -- using a countLimit of "u32bitFMT" which gets us "u32bitFMT" mers\n",
                 countLimit, numHitsAtCount[countLimit]);
#endif

      for (u32bit qi=0; qi<query->numberOfMers(); qi++) {
        if ((query->getSkip(qi) == false) &&
            (PS->get(query->getMer(qi), state->posn, state->posnMax, state->posnLen))) {
          if (state->posnLen < 3) {
            for (u32bit x=0; x<state->posnLen; x++)
              state->posn[x] += GENlo;

#ifdef SAVE_HITS_TO_FILES
            {
              for (u32bit x=0; x<state->posnLen; x++)
                fprintf(state->hitsDump, u32bitFMT" "u64bitFMT"\n", qi, state->posn[x]);
            }
#endif
            HM->addHits(qi, state->posn, state->posnLen);
          }
        }
      }

      ///////////////////////////////////////
      //
      //  Chain the hits
      //
      HM->filter(rc ? 'r' : 'f', 0.01, 0, theHits, theHitsLen, theHitsMax);

      //  Mark this hit as dead
      //
      theHits[h]._status |= AHIT_DISCARDED;

      delete HM;
      delete PS;
      delete MS;
    }
  }


#ifdef SAVE_HITS_TO_FILES
 {
   fclose(state->hitsDump);

   char  filename[1024];
   if (rc)
     sprintf(filename, "hitsDump-%08d-r.3.final", (int)seq->getIID());
   else
     sprintf(filename, "hitsDump-%08d-f.3.final", (int)seq->getIID());
   errno = 0;
   state->hitsDump = fopen(filename, "w");
   if (errno) {
     fprintf(stderr, "Failed to open '%s' for hitsDump!\n", filename);
   }

   for (u32bit x=0; x<theHitsLen; x++)
     fprintf(state->hitsDump, u32bitFMT" "u32bitFMT" "u32bitFMT" "u32bitFMT" "u32bitFMT" "u32bitFMT" %c "u32bitHEX"\n",
             theHits[x]._dsIdx,
             theHits[x]._dsLo,
             theHits[x]._dsHi,
             theHits[x]._covered,
             theHits[x]._matched,
             theHits[x]._numMers,
             theHits[x]._status & AHIT_DIRECTION_MASK ? 'f' : 'r',
             theHits[x]._status);
 }
#endif

  delete matrix;
  delete query;

#ifdef SAVE_HITS_TO_FILES
  fclose(state->hitsDump);
#endif
}




